# TTPixel - (c) 2002 Satzbau GmbH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# author: Abel Deuring - <a.deuring@satzbau-gmbh.de
# version: 0.0.2

import threading
import urllib
import string, re
import StringIO
from htmlentitydefs import entitydefs
import Image

from Persistence import Persistent
import Acquisition
from AccessControl import getSecurityManager, ClassSecurityInfo
from Globals import HTMLFile, DTMLFile, MessageDialog

from Products.ExtFile.ExtFile import ExtFile, TRANSMAP
import OFS

import freetype

_AddPermission = "Add ExtFiles" # xxx think about better permission

##########################################################################
# Enable TTPixelMethod only if you know ehat you are doing!
# Naive usage can easily lead to a high processor load,
# and DoS attacks are very easy, if TTPixelMethods are publicly
# accessible

EnableTTPixelMethod = 1

LOCAL_ENGINE = 1

manage_addTTFontForm=HTMLFile('dtml/addTTFontForm', globals())

def manage_addTTFont(self, id='', file='', REQUEST=None):
    """ Add a font to a TTFontManager instance """
    if self.this().meta_type != 'TTFontManager':
        raise "TTFont instances can only be added to TTFontManager instances!"
    # following stuff is stolen from ExtFile:
    if not id and hasattr(file, 'filename'):
        id = file.filename
        id = id[max(string.rfind(id, '/'),
                    string.rfind(id, '\\'),
                    string.rfind(id, ':')
                   )+1:]
    string.translate(id, TRANSMAP)
    self = self.this()
    
    font = TTFont(id, self, file)

    if REQUEST is not None:
        return self.manage_main(self, REQUEST)
        
class TTFont(ExtFile):

    meta_type = 'TTFont'

    security = ClassSecurityInfo()
    security.declareObjectProtected('View') #xxx not very clever
    
    # known names for NameID 2 (Subfamily) in the TT name table
    # value is: (bold, italic)
    # 
    # xxx other possible values include: 'Heavy', 'Heav', 'Medium',
    # 'Black', 'Light', 'Condensed', 'Demi', 'BookComp', 'DemiCond'
    # 'BookCond', 'Alternate', 'Antiqua', 'BookCondensed', 'Book Condensed',
    # 'Russisch' (no joke), 'Thin'
    #
    _subFamilyAttr = { 'Regular': (0, 0),
                       'Normal':  (0, 0),
                       'Roman':   (0, 0),
                       'Book':    (0, 0),
                       'Bold':    (1, 0),
                       'Italic':  (0, 1),
                       'Bold Italic': (1, 1),
                     }
    
    def __init__(self, id, container, file):
        """ id:        ID of the object
            container: the TTFontManager instance which will contain
                       this font
            file:      file with font data
        """
    
        # we don't need (yet??) title & description
        # permission_check is always 'on'
        ExtFile.__init__(self, id, '', '', 1)
        
        # the usual way to call container._setObject in manage_add...
        # and to use manage_afterAdd for further initialization
        # doesn't work with this class:
        # we want to initialize a few parameters which we must read
        # from the TTF file. So we need to store the file first; then open it
        # via the freetype engine. In order to be able to store the file,
        # ExtFile needs the location of the object in ZODB (call of absolute_url)
        # this works only after _setObject has been called.
        # So we do everything here
        
        container._setObject(id, self)
        # we need the acquisition machinery:
        container._getOb(id).manage_file_upload(file, '')
        engine = container.getTTEngine()
        self._v_face = face = freetype.TT_Open_Face(engine, self._get_filename(self.filename))
        # get some font properties
        # font name. At present, let's insist on the name as given
        # for (platform, encoding) == (1,0) (Mac Roman), since Windows
        # Unicode (3,1) gives the name as a Unicode string, by PyFT
        # returns it as a "normal" string

        # xxx am I missing somthing, or has PyFT really not a function
        # like TT_Get_Name_Count ??
        loop = 1
        index = 0
        while loop != 0:
            # useful TT font information is available at 
            # http://developer.apple.com/fonts/TTRefMan/ ,
            # especially http://developer.apple.com/fonts/TTRefMan/RM06/Chap6.html
            try:
               # Font Family
               (platformID, encodingID, languageID, nameID) = freetype.TT_Get_Name_ID(face, index)
            except:
               loop = 0
            if loop:
               if (platformID, encodingID, languageID, nameID) == (1, 0, 0, 1):
                   self.name = freetype.TT_Get_Name_String(face, index)
               elif (platformID, encodingID, languageID, nameID) == (1, 0, 0, 2):
                   # "Subfamily": bold, italic etc
                   text = freetype.TT_Get_Name_String(face, index)
                   if text in self._subFamilyAttr.keys():
                       (self.bold, self.italic) = self._subFamilyAttr[text]
               elif (platformID, encodingID, languageID, nameID) == (1, 0, 0, 4):
                   self.title = freetype.TT_Get_Name_String(face, index)
               index += 1
        
        if not hasattr(self, 'name'):
            raise ("can't find a font name in %s" % self.filename)
        
        if not hasattr(self, 'bold'):
            self.bold = 0
            self._v_msg = """TTFont created!<br>
              Warning: font attributes automatically generated"""
        if not hasattr(self, 'italic'):
            self.italic = 0
            self._v_msg = """TTFont created!<br>
              Warning: font attributes automatically generated"""
    
    security.declareProtected('View', 'getFontProperties')
    def getFontProperties(self):
        """ return (font name, bold, italic)
        """
        return (self.name, self.bold, self.italic)
    
    security.declarePrivate('TT_Open_Face')
    def get_TT_Face(self, engine):
        """ open a freetype face object """
        # xxx add caching!!!
        return freetype.TT_Open_Face(engine, self._get_filename(self.filename))

manage_addTTFontManagerForm=HTMLFile('dtml/addTTFontManagerForm', globals())

def manage_addTTFontManager(self, id, title='', REQUEST=None):
    """ Add a font manager instance """
    self = self.this()
    self._setObject(id, TTFontManager(id, title))
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

class TTFontManager(OFS.Folder.Folder):
    """ A simple font manager. Provides access to font files (used by 
        TTStyleSheet and TTPixelText); groups fonts into font families
    """
    meta_type = 'TTFontManager'
    
    #meta_types = ('TTFont',)

    security = ClassSecurityInfo()
    security.declareObjectProtected('View') #xxx not very clever
    
    TT_Engine = None

    def __init__(self, id, title=None):
        self.id = id
        self.title = title

    security.declarePrivate('getTTEngine')
    def getTTEngine(self):
        if 0:
            # xxx
            # With the current PyFT version (0.65), things can become
            # quite nasty, if we store the engine on a per-thread
            # basis. It can happen that the engine "disappers", while
            # other freetype objects like font faces or font instances 
            # still refer to it. No idea how this can happen, since
            # all the other freetype objects "live" only during a 
            # method call...
            res = getattr(self.aq_self, '_v_TTEngine', None)
            if res is None:
                res = self._v_TTEngine = freetype.TT_Init_FreeType()
            return res
        else:
            # Let's use a new TT_Engine in every call to 
            # TTStyleSheet.rasterText
            #
            # The main point is probably that freetype objects manage their
            # internal pointers to a TT_Engine, while PyFT (more precisely,
            # freetype.i) does not know about these relations. So the Zope
            # machinery may delete a TT_Engine, before all related objects
            # like TT_Face or TT_Instance are destroyed. Best fix is probably
            # to add a Python pointer to TT_Face / TT_Instance objects and
            # to write proper constructors / destructors for these classes.
            
            if TTFontManager.TT_Engine is None:
                TTFontManager.TT_Engine = freetype.TT_Init_FreeType()
            # 
            freetype.TT_Set_Raster_Gray_Palette(TTFontManager.TT_Engine, 
                                                    (0, 64, 128, 192, 255))
            
            
            return TTFontManager.TT_Engine

    security.declareProtected('View management screens', 'getFont') # xxx
    def getFont(self, font, bold=0, italic=0):
        """ return a TTFont object for the given font name or None """
        obList = self.objectValues('TTFont')
        test = (font, bold, italic)
        for ob in obList:
            if ob.getFontProperties() == test:
                return ob
        return None

    security.declareProtected('View management screens', 'getFaceNames') # xxx
    def getFaceNames(self):
        res = []
        obList = self.objectValues('TTFont')
        for ob in obList:
            (name, bold, italic) = ob.getFontProperties()
            if not name in res:
                res.append(name)
        return res
    
    security.declarePrivate('findFont')
    def findFont(self, name, bold, italic):
        obList = self.objectValues('TTFont')
        for ob in obList:
            if (name, bold, italic) == ob.getFontProperties():
                return ob
        return None
    

manage_addTTStyleSheetForm=HTMLFile('dtml/addTTStyleSheetForm', globals())

def manage_addTTStyleSheet(self, id, title='', REQUEST=None):
    """ Add a style sheet instance """
    self = self.this()
    self._setObject(id, TTStyleSheet(id, title))
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

class TTStyleSheet(OFS.SimpleItem.SimpleItem):
    meta_type = 'TTStyleSheet'

    security = ClassSecurityInfo()
    security.declareObjectProtected('View') #xxx not very clever
    
    manage_options = ({'label': 'Edit', 'action': 'manage_edit'},
                      {'label': 'Test', 'action': 'manage_test'}
                     ) + OFS.SimpleItem.SimpleItem.manage_options
    
    def __init__(self, id, title=''):
        self.id = id
        self.title = title
        self.bold=0
        self.italic=0
        self.size=12
        self.color=(0, 0, 0)
        self.bgcolor=(255, 255, 255)
        self.oversampling = 1
        self.ds_algo = Image.ANTIALIAS
        self.width = 0
        self.height = 0
        self.hAlign = 'l'
        self.vAlign = 't'
        self.autoWrap = 0
        self.hardLBreak = 0
        self.imageFormat = 'GIF'
    
    security.declareProtected('View management screens', 'manage_edit')
    manage_edit = HTMLFile('dtml/editTTStyleSheet', globals())
    
    security.declareProtected('View management screens', 'manage_test')
    manage_test = HTMLFile('dtml/testTTStyleSheet', globals())
    
    security.declareProtected('View management screens', 'getFaceName') # xxx
    def getFaceName(self):
        """ return the name of the selected font """
        return getattr(self, 'fontFace', None)

    security.declareProtected('View management screens', 'getFaceNames')
    def getFaceNames(self):
        """ return the names of the accessible fonts """
        fmList = self.superValues('TTFontManager')
        res = []
        for mgr in fmList:
            fonts = mgr.getFaceNames()
            for f in fonts:
                if not f in res:
                    res.append(f)
        return res
            
    security.declarePrivate('getFont')
    def getFont(self):
        """ return (font, fontmanager) for "our" font """
        fmList = self.superValues('TTFontManager')
        for mgr in fmList:
            font = mgr.findFont(self.fontFace, self.bold, self.italic)
            if font: return (font, mgr)
        return None
    
    security.declarePrivate('getFontID')
    def getFontID(self):
        """ return a hash value for the font or None, if the font 
            cannot be found
        """
        r = self.getFont()
        if r:
            font = r[0]
            return hash(font.name + str(font.bold) + str(font.italic) \
                        + str(self.size) + str(self.color) + str(self.bgcolor) \
                        + str(self.oversampling) \
                        + str(self.getWidth()) + str(self.getHeight()) \
                        + str(self.getPaddingLeft()) + str(self.getPaddingRight()) \
                        + str(self.getPaddingTop()) + str(self.getPaddingBottom()) \
                        + str(self.getHAlign()) + str(self.getVAlign()) \
                        + str(self.getAutoWrap()) + str(self.getHardLBreak()))
        else: return None
    
    _ds_algos = {'ANTIALIAS': Image.ANTIALIAS,
                 'BICUBIC':   Image.BICUBIC,
                 'BILINEAR':  Image.BILINEAR }
    _ds_names = {Image.ANTIALIAS: 'ANTIALIAS',
                 Image.BICUBIC:   'BICUBIC',
                 Image.BILINEAR:'BILINEAR'}
                 
    security.declarePrivate('getDsAlgo')
    def getDsAlgo(self):
        return getattr(self.aq_base, 'ds_algo', Image.ANTIALIAS)
    
    security.declareProtected('View management screens', 'getDsAlgoName')
    def getDsAlgoName(self):
        return self._ds_names[self.getDsAlgo()]
    
    security.declareProtected('View management screens', 'getImageFormat')
    def getImageFormat(self):
        return getattr(self.aq_base, 'imageFormat', 'GIF')
    
    security.declareProtected('View management screens', 'edit')
    def edit(self, fontFace, size, title, bold=0, italic=0, oversampling=0,
             ds_algo='ANTIALIAS', red=0, green=0, blue=0, 
             bg_red=255, bg_green=255, bg_blue=255, 
             width=0, height=0, 
             paddingLeft=0, paddingRight=0, paddingTop=0, paddingBottom=0,
             hAlign='l', vAlign='t', 
             autoWrap=0, hardLBreak=0, imageFormat='GIF', REQUEST=None):
        """ update the settings """
        self.fontFace = fontFace
        self.title = title
        # xxx return a more polite error message, if float() bails out
        size = float(size)
        if size <= 0:
            raise "font size must be a positive number"
        self.size = size

        if italic:
            self.italic = 1
        else:
            self.italic = 0
        if bold:
            self.bold = 1
        else:
            self.bold = 0
        
        oversampling = int(oversampling)
        if not (1 <= oversampling <= 10):
            raise "oversampling must be between 1 and 10!"
        self.oversampling = oversampling
        
        if ds_algo in self._ds_algos.keys():
            self.ds_algo = self._ds_algos[ds_algo]
        else:
            raise ("unknown downsampling algorithm: %s" % ds_algo)
        
        for (val, name) in ((red, 'red'),
                            (green, 'green'),
                            (blue, 'blue'),
                            (bg_red, 'background red'),
                            (bg_green, 'background green'),
                            (bg_blue, 'background blue')):
            try:
               x = int(val)
            except:
                raise ("%s component must be a number between 0 and 255" % name)
            if not 0 <= x <= 255:
                raise ("%s component must be a number between 0 and 255" % name)
        
        self.color=(int(red), int(green), int(blue))
        self.bgcolor = (int(bg_red), int(bg_green), int(bg_blue))
        
        self.width = int(width)
        self.height = int(height)
        self.paddingLeft = int(paddingLeft)
        self.paddingRight = int(paddingRight)
        self.paddingTop = int(paddingTop)
        self.paddingBottom = int(paddingBottom)
        self.hAlign = hAlign
        self.vAlign = vAlign
        if autoWrap:
            self.autoWrap = 1
        else:
            self.autoWrap = 0
        if hardLBreak:
            self.hardLBreak = 1
        else:
            self.hardLBreak = 0
        
        if imageFormat not in ('GIF', 'JPEG', 'PNG'):
            raise ("unknown image format: %s" % imageFormat)
        self.imageFormat = imageFormat
        
        # now invalidate the cache for all TTPixelText instances which
        # use this style sheet
        self.aq_parent.ZopeFindAndApply(self.aq_parent, 
            obj_ids=None, 
            obj_metatypes=('TTPixelText',),
            obj_searchterm=None, 
            obj_expr=None,
            obj_mtime=None, 
            obj_mspec=None,
            obj_permission=None, 
            obj_roles=None,
            search_sub=0,
            REQUEST=None, 
            result=None, 
            pre='',
            apply_func=lambda ob, path, style=self.id: _pixelUnCache(ob, path, style),
            apply_path='')
        
        if REQUEST:
            return self.manage_edit(self, REQUEST)

    security.declareProtected('View management screens', 'test')
    def test(self):
        raise "Test is not yet implemented!"

    security.declareProtected('View management screens', 'getWidth')
    def getWidth(self):
        return getattr(self, 'width', 0)

    security.declareProtected('View management screens', 'getPaddingLeft')
    def getPaddingLeft(self):
        return getattr(self, 'paddingLeft', 0)

    security.declareProtected('View management screens', 'getPaddingRight')
    def getPaddingRight(self):
        return getattr(self, 'paddingRight', 0)

    security.declareProtected('View management screens', 'getPaddingTop')
    def getPaddingTop(self):
        return getattr(self, 'paddingTop', 0)

    security.declareProtected('View management screens', 'getPaddingBottom')
    def getPaddingBottom(self):
        return getattr(self, 'paddingBottom', 0)

    security.declareProtected('View management screens', 'getHeight')
    def getHeight(self):
        return getattr(self, 'height', 0)

    security.declareProtected('View management screens', 'getHAlign')
    def getHAlign(self):
        res = getattr(self, 'hAlign', 'l')
        try:
            if not res in 'lcr':
                return 'l'
        except:
            return 'l'
        return res

    security.declareProtected('View management screens', 'getVAlign')
    def getVAlign(self):
        res = getattr(self, 'vAlign', 't')
        if not res in 'tmb':
            return 't'
        return res

    security.declareProtected('View management screens', 'getAutoWrap')
    def getAutoWrap(self):
        return getattr(self, 'autoWrap', 0)

    security.declareProtected('View management screens', 'getHardLBreak')
    def getHardLBreak(self):
        return getattr(self, 'hardLBreak', 0)

    _lineRe = re.compile('\r\n|\n\r|\r|\n')
    _spRe   = re.compile(r'\s+')

    security.declarePrivate('renderText')
    def renderText(self, text):
        """ render the given text. Returns a PIL image object
            mainly copied from PyFT.freetype.drawString
        """
        face = ins = None
        
        try:
            text = _resolveEntities(text)
            try:
                text = string.expandtabs(text)
            except TypeError:
                pass
            
            if self.getHardLBreak():
                lines = self._lineRe.split(text)
            else:
                lines = [text]
            for i in range(len(lines)):
                lines[i] = self._spRe.split(lines[i])
            
            font, mgr = self.getFont()
            if LOCAL_ENGINE:
                e = freetype.TT_Init_FreeType()
            else:
                e = mgr.getTTEngine()

            face = font.get_TT_Face(e)

            props = freetype.TT_Get_Face_Properties(face)

            ins = freetype.TT_New_Instance(face)
            
            oversampling = self.oversampling

            freetype.TT_Set_Instance_Resolutions(ins, 
              96 * oversampling, 96 * oversampling)
            freetype.TT_Set_Instance_PointSize(ins, self.size)

            for id in range(props.num_CharMaps):
                tup = freetype.TT_Get_CharMap_ID(face, id)
                if tup == (3,1):
                    # Windows Unicode mapping
                    cm = freetype.TT_Get_CharMap(face, id)
                    break
                # xxx raise an error, if no charmap found

            glyphCache = {' ': freetype.TT_New_Glyph(face) }
            freetype.TT_Load_Glyph(ins, glyphCache[' '], 
                                   freetype.pTT_Char_Index(cm, ' '),
                                   freetype.TTLOAD_DEFAULT)
            metricCache = {' ': freetype.TT_Get_Glyph_Metrics(glyphCache[' '])}
            spaceWidth26F6 = metricCache[' '].advance
            
            for k in range(len(lines)):
              line = lines[k]
              lWidth26F6 = 0
              for j in range(len(line)):
                word = line[j]
                wWidth26F6 = 0
                for cc in word:
                  if not glyphCache.has_key(cc):
                    glyphCache[cc] = freetype.TT_New_Glyph(face)
                    # xxx using pTT_CHar_Index(cm, cc) does not work,
                    # if ord(cc) >= 128
                    freetype.TT_Load_Glyph(ins, glyphCache[cc], 
                                           freetype.TT_Char_Index(cm, ord(cc)), 
                                           freetype.TTLOAD_DEFAULT)
                    metricCache[cc] = freetype.TT_Get_Glyph_Metrics(glyphCache[cc])
                  wWidth26F6 += metricCache[cc].advance
                line[j] = (word, wWidth26F6)
                lWidth26F6 += wWidth26F6
                if j:
                    lWidth26F6 += spaceWidth26F6
              lines[k] = (line, lWidth26F6)

            metrics = freetype.TT_Get_Instance_Metrics(ins)
            maxyMax26F6 = (1L * props.horizontal.Ascender * metrics.y_scale) >> 16
            minyMin26F6 = (1L * props.horizontal.Descender * metrics.y_scale) >> 16
            lineGap26F6 = (1L * props.horizontal.Line_Gap * metrics.y_scale) >> 16
            
            imWidth = self.getWidth()
            imHeight = self.getHeight()
            # set a minimum margin of 1 pixel. This avoids "cut off" char
            # images.
            paddingLeft = self.getPaddingLeft()
            paddingRight = self.getPaddingRight()
            paddingTop = self.getPaddingTop()
            paddingBottom = self.getPaddingBottom()

            if paddingLeft <= 0:
                paddingLeft = 1
            if paddingRight <= 0:
                paddingRight = 1
            if paddingTop <= 0:
                paddingTop = 1
            if paddingBottom <= 0:
                paddingBottom = 1
                
            if imWidth > 0:
                allowedLWidth26F6 = 64 * oversampling \
                                  * (imWidth - paddingLeft - paddingRight)
            else:
                allowedLWidth26F6 = 0
            maxLineWidth26F6 = 0
            lines2 = []
            # check, if all lines fit within the given image size
            for (line, lWidth26F6) in lines:
                if allowedLWidth26F6 > 0 and lWidth26F6 > allowedLWidth26F6:
                    line2 = []
                    lWidth26F6 = -spaceWidth26F6
                    while line:
                        (word, wWidth26F6) = line.pop(0)
                        # at least one word per line; wrap text into next line,
                        # if automatic line break is enabled
                        if len(line2) == 0 or not self.autoWrap or \
                           lWidth26F6 + wWidth26F6 + spaceWidth26F6 < allowedLWidth26F6:
                            line2.append((word, wWidth26F6))
                            lWidth26F6 += wWidth26F6 + spaceWidth26F6
                        elif len(line2):
                            lines2.append((line2, lWidth26F6))
                            line2 = [(word, wWidth26F6)]
                            lWidth26F6 = wWidth26F6
                    if line2:
                        lines2.append((line2, lWidth26F6))
                else:
                    lines2.append((line, lWidth26F6))
                maxLineWidth26F6 = max(maxLineWidth26F6, lWidth26F6)
            textHeight26F6 = len(lines2) * (maxyMax26F6 - minyMin26F6 + lineGap26F6) \
                           - lineGap26F6

            if imWidth <= 0:
                imWidth = width = int((int((maxLineWidth26F6+63)/64) + oversampling-1)/oversampling) \
                                + paddingLeft + paddingRight

            if imHeight <= 0:
                imHeight = int((int((textHeight26F6 + 63) / 64) +oversampling-1)/oversampling)\
                         + paddingTop + paddingBottom
            
            hAlign = self.getHAlign()
            vAlign = self.getVAlign()
            
            freetype.TT_Set_Raster_Gray_Palette(e, (0, 64, 128, 192, 255))

            res = Image.new('RGB', (imWidth * oversampling, 
                                    imHeight * oversampling), 
                                    self.bgcolor)
            rasterSizeH = (imWidth - paddingLeft - paddingRight + 2) * oversampling
            rasterSizeV = (imHeight - paddingTop - paddingBottom + 2) * oversampling
            pasteOffset = ((paddingLeft - 1) * oversampling, 
                           (paddingTop - 1) * oversampling)
            if vAlign == 'b':
                curY26F6 = textHeight26F6 - maxyMax26F6 \
                         + 64 * oversampling
            elif vAlign == 'm':
                curY26F6 = rasterSizeV * 32  + textHeight26F6 / 2 \
                         - maxyMax26F6
            else:
                curY26F6 = (rasterSizeV-1) * 64 \
                         - maxyMax26F6
            
            # TT_Raster_Map is not properly cleared!
            raster = freetype.TT_Raster_Map(rasterSizeH, rasterSizeV, 'pix')
            for (line, lWidth26F6) in lines2:
              if hAlign == 'r':
                  #curX26F6 = (imWidth - paddingRight) * oversampling * 64 - lWidth26F6
                  curX26F6 = (rasterSizeH-oversampling) * 64 - lWidth26F6
              elif hAlign == 'c':
                  #curX26F6 = imWidth * 32 *oversampling - lWidth26F6 / 2
                  curX26F6 = rasterSizeH * 32 - lWidth26F6 / 2
              else:
                  #curX26F6 = 64 * oversampling * paddingLeft
                  curX26F6 = 64 * oversampling
              for (word, wWidth26F6) in line:
                for char in word:
                    raster = freetype.TT_Raster_Map(rasterSizeH, 
                                                    rasterSizeV, 'pix')
                    freetype.TT_Get_Glyph_Pixmap(glyphCache[char], raster, 
                                        curX26F6, curY26F6)
                    mask = freetype.toMask(raster, 'pix')
                    res.paste(self.color, pasteOffset, mask)
                    curX26F6 = curX26F6 + metricCache[char].advance
                curX26F6 += spaceWidth26F6
              curY26F6 -= maxyMax26F6 - minyMin26F6 + lineGap26F6

            if oversampling != 1:
                # ANTIALIAS does not work for PIL 1.1.2 and older versions
                #res = res.resize((imWidth,imHeight), Image.ANTIALIAS)
                #res = res.resize((imWidth,imHeight), Image.BICUBIC)
                #res = res.resize((imWidth,imHeight), Image.BILINEAR)
                res = res.resize((imWidth,imHeight), self.getDsAlgo())
            
        finally:
            if ins != None:
                freetype.TT_Done_Instance(ins)
            if face != None:
                freetype.TT_Close_Face(face)
        return res
            

def _pixelUnCache(ob, path, style):
    if ob.stylesheet == style:
        ob.ZCacheable_invalidate()

def _getIndices(cm, text):
    # xxx replacement for PyTT_STring_Indices
    # xxx should be moved into class TTFont; result should be cached
    x = freetype.TT_CharMap_First(cm)
    res = []
    tmp = {}
    while x:
        tmp[x[0]] = x[1]
        try:
            # TT_Char_Map_Next bails out when the last symbol is reached
            x = freetype.TT_CharMap_Next(cm, x[0])
        except:
            x = None
    for c in text:
        res.append(tmp[ord(c)])
    return res
        

manage_addTTPixelTextForm=HTMLFile('dtml/addTTPixelTextForm', globals())

def manage_addTTPixelText(self, id, title='', stylesheet='', REQUEST=None):
    """ Add a pixeltext instance """
    self = self.this()
    self._setObject(id, TTPixelText(id, title, self, stylesheet))
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

class TTPixelText(OFS.Image.Image):
    meta_type = 'TTPixelText'

    security = ClassSecurityInfo()
    security.declareObjectProtected('View')
    
    security.declareProtected('View management screens', 'manage_editForm')
    security.declareProtected('View management screens', 'manage_main')
    security.declareProtected('View management screens', 'manage')
    manage_editForm = HTMLFile('dtml/editTTPixelTextForm', globals())
    manage_editForm._setName('manage_editForm')
    manage = manage_main = manage_editForm
    
    manage_options =  (({'label':'Edit', 'action':'manage_main',
                       'help':('OFSP','Image_Edit.stx')},
                       )                     
                       + OFS.Image.Image.manage_options[1:]
                      )
    
    def __init__(self, id, title, container, stylesheet=''):
        OFS.Image.Image.__init__(self, id, '', '')
        self.title = title
        self.stylesheet = stylesheet
        if stylesheet:
            self.__of__(container)._makeImage()
    
    security.declareProtected('View', 'view_image_or_file')
    view_image_or_file =HTMLFile('dtml/imageView',globals())
    
    security.declareProtected('Change Images and Files', 'quotedHTML')
    def quotedHTML(self,
                   text=None,
                   character_entities=(
                       (('&'), '&amp;'),
                       (("<"), '&lt;' ),
                       ((">"), '&gt;' ),
                       (('"'), '&quot;'))): #"
        """ HTML-quote special characters """
        # stolen from DocumentTemplate/DT_HTML.py
        if text is None: text=self.read_raw()
        for re,name in character_entities:
            if string.find(text, re) >= 0:
                text=string.join(string.split(text,re),name)
        return text

    security.declareProtected('Change Images and Files', 'edit')
    def edit(self, title, stylesheet, REQUEST=None):
        """ edit """
        if self.title != title or self.stylesheet != stylesheet:
            self.title = title
            self.stylesheet = stylesheet
            self._makeImage()
        if REQUEST:
            return self.manage_main(self, REQUEST)
    
    security.declareProtected('Add Documents, Images, and Files', 'manage_upload')
    def manage_upload(self, file='', REQUEST=None):
        """ override Image.manage_upload: that's no reasonable for us
        """
        raise "can't upload a TTPixelText object"

    def _makeImage(self):
        """ render the title """
        style = getattr(self, self.stylesheet, None)
        if style is None:
            raise "can't find the stylesheet", self.stylesheet
        img = style.renderText(self.title)
        self.styleID = style.getFontID()
        f = StringIO.StringIO()
        img.save(f, style.getImageFormat())
        OFS.Image.Image.manage_upload(self, f)
    
    
    def _testUpdate(self):
        style = getattr(self, self.stylesheet, None)
        if style is None:
            raise "can't find the stylesheet", self.stylesheet
        # doesn't work with TTPixelMethod, because instances of TTPixelText
        # used there never have the "_p_mtime" attribute: they are not
        # stored in the ZODB!
        # if not self._p_mtime or long(self._p_mtime) <= long(style._p_mtime):
        #     self._makeImage()
        test = style.getFontID()
        styleID = getattr(self, 'styleID', None)
        if styleID != test:
            self._makeImage()
        
        
    
    security.declareProtected('View', 'index_html')
    def index_html(self, REQUEST, RESPONSE):
        """ return the image contents. 
            Check, if the style sheet has been changed; if yes,
            render the image again
        """
        self._testUpdate()
        return OFS.Image.Image.index_html(self, REQUEST, RESPONSE)
    
    security.declareProtected('View', 'tag')
    def tag(self, height=None, width=None,
            scale=0, xscale=0, yscale=0, css_class=None, **args):
        """ return the <img ..> tag. Check, if the style sheet has
            been changed; if yes, render the image again.
        """
        self._testUpdate()
        return apply(OFS.Image.Image.tag, 
                 (self, height, width, self.title, scale, xscale, yscale, 
                  css_class), args)


if EnableTTPixelMethod:
    manage_addTTPixelMethodForm=HTMLFile('dtml/addTTPixelMethodForm', globals())

    def manage_addTTPixelMethod(self, id, title='', file='', stylesheet='', REQUEST=None):
        """ Add a pixeltext instance 
            xxx add parameter 'file' to the add form!
        """
        self = self.this()
        id = str(id)
        title = str(title)
        if type(file) is not type(''): file = file.read()
        self._setObject(id, TTPixelMethod(id, title, file, stylesheet))
        if REQUEST is not None:
            return self.manage_main(self, REQUEST)


    class TTPixelMethod(OFS.DTMLMethod.DTMLMethod):
        """ renders text like a DTMLMethod, but returns an <img ...> tag 
            which "points" to a dynamically generated image with the
            rendered text
            
            A call to the index_html method of this class returns this image
        """

        # the followng is a common cache for all instances of TTPixelMethod!
        _cache = {}
        _cacheUsed = []
        _cacheSize = 500
        _cacheLock = threading.Lock()

        meta_type = 'TTPixelMethod'

        security = ClassSecurityInfo()
        security.declareObjectProtected('View')
        
        security.declareProtected('Change DTML Methods', 'manage_editForm')
        security.declareProtected('Change DTML Methods', 'manage')
        security.declareProtected('Change DTML Methods', 'manage_main')
        security.declareProtected('Change DTML Methods', 'manage_editDocument')

        
        manage_editForm=DTMLFile('dtml/documentEdit', globals())
        manage_editForm._setName('manage_editForm')
        manage=manage_main=manage_editDocument=manage_editForm
        
        
        manage_options = ({'label':'Edit', 'action':'manage_main',
                          'help':('OFSP','DTML-DocumentOrMethod_Edit.stx')},
                         ) + OFS.DTMLMethod.DTMLMethod.manage_options[1:]
        
        def __init__(self, id, title='', file='', stylesheet=''):
            OFS.DTMLMethod.DTMLMethod.__init__(self, file, __name__=id)
            self.stylesheet = stylesheet
            self.title = title
        
        
        security.declareProtected('Change DTML Methods', 'manage_edit')
        def manage_edit(self,data,title,stylesheet,SUBMIT='Change',
                        dtpref_cols='50',dtpref_rows='20',REQUEST=None):
            """ same as DTMLMethod.manage_edit, except that we have the
                additional parameter stylesheet
            """
            if SUBMIT == 'Change' or SUBMIT == 'Save Changes':
                self.stylesheet = stylesheet
            return OFS.DTMLMethod.DTMLMethod.manage_edit(self,data,title,
                       SUBMIT,dtpref_cols,dtpref_rows,REQUEST)
            

        url_re = re.compile('src=".*?"')
        def __call__(self, client, REQUEST, RESPONSE=None, **kw):
            """ get the text to be "pixel-rendered" by calling 
                DTMLMethod.__call__ 
                keep this text as a an "URL appendix" and return
                an URL for the image which represents the acquisition path
                + img method + the rendered text
            """
            text = apply(OFS.DTMLMethod.DTMLMethod.__call__, 
                         (self, client, REQUEST, RESPONSE), kw)
            
            chain = Acquisition.aq_chain(self, 0)
            
            # we need only the 'real objects' from the acquisition path
            if chain[-1].__class__.__name__ == 'RequestContainer':
                chain = chain[:-1]

            path = []
            for ob in chain:
                if callable(ob.id): path.append(ob.id())
                else:               path.append(ob.id)
            path.reverse()
            
            utext = urllib.quote(text)
            # encode more characters that might "confuse" the server
            # when parsing the URL, or which might make it difficult to
            # save a file copy of an HTML page with images
            utext = utext.replace('/', '%2F')
            utext = utext.replace('&', '%26')
            utext = utext.replace(';', '%3B')
            utext = utext.replace('Q', '%51')
            utext = utext.replace('%', 'Q')
            
            img_id = 'img_' + str(self._fontID()) + '_' + utext

            try:
                path = REQUEST['SERVER_URL'] \
                    + string.join(path, '/') + '/' + img_id
            except:
                path = self.REQUEST['SERVER_URL'] \
                    + string.join(path, '/') + '/' + img_id
            
            image = self._getCache(img_id)
            if image is None:
                image = TTPixelText(img_id, text, self, 
                                    self.stylesheet)
                self._insCache(img_id, image)
            res = image.__of__(self).tag()
            return self.url_re.sub('src="'+path+'"', res)
        
        def _fontID(self):
            style = getattr(self, self.stylesheet, None)
            if style is None:
                raise "can't find the stylesheet", self.stylesheet
            return style.getFontID()

        def __getitem__(self, key):
            if key[0:4] == 'img_':
                image = self._getCache(key)
                if image is None:
                    text = key[4:]
                    text = text[text.find('_')+1:]
                    text = text.replace('Q', '%')
                    image = TTPixelText(key, urllib.unquote(text), self, self.stylesheet)
                    self._insCache(key, image)
                
                return image.__of__(self)
                        
            else: 
                raise KeyError, key
        
        def _insCache(self, img_id, image):
            self._cacheLock.acquire()
            try:
                cache = self._cache
                used = self._cacheUsed
                if cache.has_key(img_id):
                    # should not happen, but let's be sure...
                    del cache[img_id]
                    used.pop(used.index(img_id))
                elif len(cache) >= self._cacheSize:
                    id = used.pop()
                    del cache[id]
                cache[img_id] = image
                used.insert(0, img_id)
            finally:
                self._cacheLock.release()
            
        
        def _getCache(self, img_id):
            self._cacheLock.acquire()
            try:
                cache = self._cache
                if not cache.has_key(img_id): return None
                used = self._cacheUsed
                i = used.index(img_id)
                if i:
                    used.pop(i)
                    used.insert(0, img_id)
                return cache[img_id]
            finally:
                self._cacheLock.release()


_entRef = re.compile('&(.*?);')

def _resolveEntities(text):
    """ convert HTML char entities into "regular" text 
    """
    l = _entRef.split(text)
    res = ''
    while len(l) > 1:
        res += l.pop(0)
        ent = l.pop(0)
        if entitydefs.has_key(ent):
            res += entitydefs[ent]
        else:
            try:
                i = int(ent)
                res += chr(i)
            except:
                res += '&' + ent + ';'
    return res + l[0]
